# High-Level Brainfuck
#
# Copyright (c) 2004 Marek Materzok
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import parser

class CompileError(Exception):
	def __init__(self, msg = ""):
		Exception.__init__(self, "Compile error: " + msg)

class InternalCompilerError(Exception):
	def __init__(self, descr = ""):
		Exception.__init__(self, "Bad operation: " + descr)

class AlreadyFinished(InternalCompilerError):
	def __init__(self):
		InternalCompilerError.__init__(self, "already finished")

# blok kodu do systemu selektorow
class Block(object):
	def __init__(self, bno, fno, spos = 1):
		self.code = ""
		self.start_pos = spos
		self.end_pos = spos
		self.block_no = bno
		self.fun_no = fno
		self.finished = False
		self.loops = []

	def is_finished(self):
		return self.finished

	def get_code(self):
		return self.code

	def get_block_no(self):
		return self.block_no

	def shift(self, sz):
		if self.finished: raise AlreadyFinished()
		if sz >= 0: self.code += ">" * sz
		else: self.code += "<" * (-sz)

	def go_to(self, pos):
		if self.finished: raise AlreadyFinished()
		steps = pos - self.end_pos
		if steps > 0: self.code += ">" * steps
		else: self.code += "<" * (-steps)
		self.end_pos = pos

	def add(self, n):
		if self.finished: raise AlreadyFinished()
		if n < 0: 
			op = "-"
			n = -n
		elif n > 0: op = "+"
		else: return
		self.code += op * n
	
	def sub(self, n): self.add(-n)

	def go_add(self, pos, n):
		self.go_to(pos)
		self.add(n)

	def go_sub(self, pos, n):
		self.go_to(pos)
		self.sub(n)

	def start_loop(self):
		if self.finished: raise AlreadyFinished()
		self.code += "["
		self.loops.append(self.end_pos)

	def end_loop(self):
		if self.finished: raise AlreadyFinished()
		if len(self.loops) == 0:
			raise InternalCompilerError("ending not started loop")
		p = self.loops.pop()
		if p != self.end_pos: 
			raise InternalCompilerError("loop didn't end on starting pos")
		self.code += "]"

	def input(self):
		if self.finished: raise AlreadyFinished()
		self.code += ","

	def output(self):
		if self.finished: raise AlreadyFinished()
		self.code += "."

	def ext_testloop(self, src, add_dst = [], sub_dst = []):
		if self.finished: raise AlreadyFinished()
		self.go_to(src)
		self.start_loop()
		self.start_loop()
		self.sub(1)
		self.end_loop()
		for pos in add_dst:
			self.go_to(pos)
			self.add(1)
		for pos in sub_dst:
			self.go_to(pos)
			self.sub(1)
		self.go_to(src)
		self.end_loop()

	def ext_moveloop(self, src, add_dst = [], sub_dst = []):
		if self.finished: raise AlreadyFinished()
		self.go_to(src)
		self.start_loop()
		self.sub(1)
		for pos in add_dst:
			self.go_to(pos)
			self.add(1)
		for pos in sub_dst:
			self.go_to(pos)
			self.sub(1)
		self.go_to(src)
		self.end_loop()

	def finish_return(self):
		if self.finished: raise AlreadyFinished()
		self.go_to(1)
		self.finished = True

	def finish_gotoblock(self, bn):
		if self.finished: raise AlreadyFinished()
		self.go_to(0)
		self.add(bn)
		self.go_to(1)
		self.finished = True

	def finish_callfunc(self, p, fn, bn):
		if self.finished: raise AlreadyFinished()
		self.go_to(p)
		self.add(bn)
		self.go_to(p+3)
		self.add(self.fun_no)
		self.go_to(p+6)
		self.add(1)
		self.go_to(p+9)
		self.add(fn)
		self.go_to(p+12+1)
		self.finished = True

class Type(object):
	def get_size(self): pass
	def create(self): pass
	def __eq__(self, obj): pass
	def __ne__(self, obj): return not self.__eq__(obj)

class ByteType(Type):
	def get_size(self): return 2
	def create(self, pos = 0): return Byte(pos)
	def __eq__(self, obj):
		if type(obj) == ByteType: return True
		else: return False

class VoidType(Type):
	def get_size(self): return 0
	def create(self, pos = 0): return Void(pos)
	def __eq__(self, obj):
		if type(obj) == VoidType: return True
		else: return False

class ArrayType(Type):
	def __init__(self, alen, vtype):
		Type.__init__(self)
		self._len = alen
		self._vtype = vtype
		self._ctype = ByteType()
		self._cellsize = vtype.get_size() * 2 + self._ctype.get_size()
		self._size = (alen + 1) * self._cellsize
	def get_size(self): return self._size
	def create(self, pos = 0): return Array(pos, self)
	def __eq__(self, obj):
		if type(obj) != ArrayType: return False
		if obj._len == self._len and obj._vtype == self._vtype: return True
		else: return False

	def get_len(self): return self._len
	def get_cellsize(self): return self._cellsize
	def get_vtype(self): return self._vtype

class Variable(object):
	def __init__(self, pos, vtype):
		self.position = pos
		self.type = vtype
	def clone(self): return self.type.create(self.position)
	def get_size(self): return self.type.get_size()
	def get_type(self): return self.type
	def get_position(self): return self.position
	def set_position(self, pos): self.position = pos
	
	# czysci zmienna
	def op_clear(self, bl): pass
	# nadpisuje zmienna kopia argumentu
	def op_copy(self, arg, bl): pass
	# nadpisuje zmienna argumentem (niszczy go)
	def op_move(self, arg, bl): pass
	# operatory destruktywne (niszcza argument)
	# operuja na stosie temp, argument jest na szczycie
	def op_add(self, arg, bl): pass
	def op_sub(self, arg, bl): pass
	def op_mul(self, arg, bl): pass
	def op_div(self, arg, bl): pass
	def op_mod(self, arg, bl): pass
	def op_eq(self, arg, bl): pass
	def op_neq(self, arg, bl): pass
	def op_and(self, arg, bl): pass
	def op_or(self, arg, bl): pass
	def op_index(self, arg, bl): pass

	def op_bool(self, bl): pass
	def op_not(self, bl): pass

# zmienna Byte
# zajmuje 2 komorki, druga jest uzywana jako tymczasowa
class Byte(Variable):
	def __init__(self, pos = 0):
		Variable.__init__(self, pos, ByteType())

	def op_clear(self, bl):
		bl.go_to(self.position)
		bl.start_loop()
		bl.sub(1)
		bl.end_loop()
	def op_copy(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		self.op_clear(bl)
		# skopiuj na miejsce i do miejsca tymczasowego
		bl.ext_moveloop(arg.position, [arg.position + 1, self.position])
		# przenies z miejsca tymczasowego z powrotem
		bl.ext_moveloop(arg.position + 1, [arg.position])
	def op_move(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		self.op_clear(bl)
		# przenies na miejsce, destruktywnie
		bl.ext_moveloop(arg.position, [self.position])
	def op_add(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		bl.ext_moveloop(arg.position, [self.position])
		return self
	def op_sub(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		bl.ext_moveloop(arg.position, [], [self.position])
		return self
	def op_mul(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		bl.ext_moveloop(self.position, [self.position + 1, arg.position + 1])
		bl.go_to(arg.position)
		bl.start_loop()
		bl.  sub(1)
		bl.  ext_moveloop(self.position + 1, [self.position])
		bl.  ext_moveloop(arg.position + 1, [self.position + 1, arg.position + 2])
		bl.  ext_moveloop(arg.position + 2, [arg.position + 1])
		bl.  go_to(arg.position)
		bl.end_loop()
		bl.ext_moveloop(arg.position + 1)
		bl.ext_moveloop(self.position + 1)
		return self
	def op_div(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		bl.ext_moveloop(self.position, [self.position + 1])
		bl.go_to(self.position + 1)
		bl.start_loop()
		self._common_divmod(arg, bl)
		bl.  go_add(arg.position + 2, 1)
		bl.  ext_testloop(arg.position + 1, [], [arg.position + 2])
		bl.  go_to(arg.position + 2)
		bl.  start_loop()
		bl.    sub(1)
		bl.    go_add(self.position, 1)
		bl.    go_to(arg.position + 2)
		bl.  end_loop()
		bl.  go_to(self.position + 1)
		bl.end_loop()
		bl.ext_moveloop(arg.position)
		return self
	def op_mod(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		bl.ext_moveloop(self.position, [self.position + 1])
		bl.go_to(self.position + 1)
		bl.start_loop()
		self._common_divmod(arg, bl)
		bl.  ext_moveloop(self.position, [arg.position + 2, arg.position + 3])
		bl.  ext_moveloop(arg.position + 2, [self.position])
		bl.  go_add(arg.position + 2, 1)
		bl.  ext_testloop(arg.position + 3, [arg.position + 2])
		bl.  go_to(arg.position + 2)
		bl.  start_loop()
		bl.    sub(1)
		bl.    ext_moveloop(self.position)
		bl.    ext_moveloop(arg.position, [arg.position + 2, self.position])
		bl.    ext_moveloop(arg.position + 2, [arg.position])
		bl.    ext_moveloop(arg.position + 1, [arg.position + 2, arg.position + 3])
		bl.    ext_moveloop(arg.position + 3, [arg.position + 1])
		bl.    ext_moveloop(arg.position + 2, [], [self.position])
		bl.    go_to(arg.position + 2)
		bl.  end_loop()
		bl.  go_add(arg.position + 2, 1)
		bl.  ext_testloop(arg.position + 1, [], [arg.position + 2])
		bl.  go_to(arg.position + 2)
		bl.  start_loop()
		bl.    sub(1)
		bl.    ext_moveloop(self.position)
		bl.    go_to(arg.position + 2)
		bl.  end_loop()
		bl.  go_to(self.position + 1)
		bl.end_loop()
		bl.ext_moveloop(arg.position)
		return self
	def _common_divmod(self, arg, bl):
		bl.  ext_moveloop(arg.position, [arg.position + 1, arg.position + 2, arg.position + 3])
		bl.  ext_moveloop(arg.position + 3, [arg.position])
		bl.  go_to(arg.position + 2)
		bl.  start_loop()
		bl.    ext_moveloop(arg.position + 2)
		bl.    go_sub(arg.position + 1, 1)
		bl.    go_sub(self.position + 1, 1)
		bl.    ext_moveloop(arg.position + 1, [arg.position + 2, arg.position + 3])
		bl.    ext_moveloop(arg.position + 3, [arg.position + 1])
		bl.    ext_moveloop(self.position + 1, [arg.position + 3, arg.position + 4])
		bl.    ext_moveloop(arg.position + 3, [self.position + 1])
		bl.    go_add(arg.position + 3, 2)
		bl.    ext_testloop(arg.position + 2, [], [arg.position + 3])
		bl.    ext_testloop(arg.position + 4, [], [arg.position + 3])
		bl.    go_add(arg.position + 2, 1)
		bl.    ext_testloop(arg.position + 3, [], [arg.position + 2])
		bl.    go_to(arg.position + 2)
		bl.  end_loop()
	def op_more(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		self._common_relation(arg, bl)
		bl.ext_moveloop(arg.position + 1)
		bl.ext_testloop(self.position + 1, [self.position])
		return self
	def op_moreeq(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		self._common_relation(arg, bl)
		bl.go_add(self.position, 1)
		bl.ext_moveloop(self.position + 1)
		bl.ext_testloop(arg.position + 1, [], [self.position])
		return self
	def op_less(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		self._common_relation(arg, bl)
		bl.ext_moveloop(self.position + 1)
		bl.ext_testloop(arg.position + 1, [self.position])
		return self
	def op_lesseq(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		self._common_relation(arg, bl)
		bl.go_add(self.position, 1)
		bl.ext_moveloop(arg.position + 1)
		bl.ext_testloop(self.position + 1, [], [self.position])
		return self
	def _common_relation(self, arg, bl):
		bl.go_add(arg.position + 2, 1)
		bl.start_loop()
		bl.  sub(1)
		bl.  go_add(arg.position + 2, 2)
		bl.  ext_moveloop(self.position, [self.position + 1, arg.position + 3])
		bl.  ext_moveloop(self.position + 1, [self.position])
		bl.  ext_testloop(arg.position + 3, [], [arg.position + 2])
		bl.  ext_moveloop(arg.position, [arg.position + 1, arg.position + 3])
		bl.  ext_moveloop(arg.position + 1, [arg.position])
		bl.  ext_testloop(arg.position + 3, [], [arg.position + 2])
		bl.  go_add(arg.position + 3, 1)
		bl.  ext_testloop(arg.position + 2, [], [arg.position + 3])
		bl.  go_to(arg.position + 3)
		bl.  start_loop()
		bl.    sub(1)
		bl.    go_add(arg.position + 2, 1)
		bl.    go_sub(self.position, 1)
		bl.    go_sub(arg.position, 1)
		bl.    go_to(arg.position + 3)
		bl.  end_loop()
		bl.  go_to(arg.position + 2)
		bl.end_loop()
		bl.ext_testloop(self.position, [self.position + 1])
		bl.ext_testloop(arg.position, [arg.position + 1])
	def op_eq(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		ret = self.op_sub(arg, bl)
		ret = self.op_not(bl)
		return ret
	def op_neq(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		ret = self.op_sub(arg, bl)
		ret = ret.op_bool(bl)
		return ret
	def op_and(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		bl.go_to(self.position + 1)
		bl.add(2)
		bl.ext_testloop(self.position, [], [self.position + 1])
		bl.ext_testloop(arg.position, [], [self.position + 1])
		bl.go_to(self.position)
		bl.add(1)
		bl.ext_testloop(self.position + 1, [], [self.position])
		return self
	def op_or(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		bl.ext_testloop(self.position, [self.position + 1])
		bl.ext_testloop(arg.position, [self.position + 1])
		bl.ext_testloop(self.position + 1, [self.position])
		return self
	def op_bool(self, bl):
		bl.ext_testloop(self.position, [self.position + 1])
		bl.ext_moveloop(self.position + 1, [self.position])
		return self
	def op_not(self, bl):
		bl.go_to(self.position + 1)
		bl.add(1)
		bl.ext_testloop(self.position, [], [self.position + 1])
		bl.ext_moveloop(self.position + 1, [self.position])
		return self

class Void(Variable):
	def __init__(self, pos = 0):
		Variable.__init__(self, pos, VoidType())

class Array(Variable):
	def __init__(self, pos, vtype):
		Variable.__init__(self, pos, vtype)
		self._cellsize = self.type.get_cellsize()
		self._vtype = self.type.get_vtype()
	def _gen_helpers(self):
		self._holder = self._vtype.create(self.position)
		self._pholder = self._vtype.create(self.position - self._cellsize)
		self._data = self._vtype.create(self.position + self._vtype.get_size())
		self._counter = ByteType().create(self.position + self._vtype.get_size() * 2)
	def op_clear(self, bl):
		self._gen_helpers()
		data = self._data.clone()
		counter = self._counter.clone()
		bl.shift(self._cellsize)
		data.op_clear(bl)
		bl.go_add(counter.position, self.type.get_len() - 1)
		bl.start_loop()
		bl.  ext_moveloop(counter.position, [counter.position + self._cellsize])
		bl.  go_add(counter.position, 1)
		bl.  shift(self._cellsize)
		data.op_clear(bl)
		bl.  go_sub(counter.position, 1)
		bl.end_loop()
		bl.go_add(counter.position, 1)
		bl.start_loop()
		bl.  go_sub(counter.position, 1)
		bl.  shift(-self._cellsize)
		bl.  go_to(counter.position)
		bl.end_loop()
	def op_copy(self, arg, bl):
		if arg.get_type() != self.type: raise CompileError("bad type")
		self._gen_helpers()
		arg._gen_helpers()
		sdata = self._data.clone()
		odata = arg._data.clone()
		counter = self._counter.clone()
		bl.shift(self._cellsize)
		sdata.op_copy(odata, bl)
		bl.go_add(counter.position, self.type.get_len() - 1)
		bl.start_loop()
		bl.  ext_moveloop(counter.position, [counter.position + self._cellsize])
		bl.  go_add(counter.position, 1)
		bl.  shift(self._cellsize)
		sdata.op_copy(odata, bl)
		bl.  go_sub(counter.position, 1)
		bl.end_loop()
		bl.go_add(counter.position, 1)
		bl.start_loop()
		bl.  go_sub(counter.position, 1)
		bl.  shift(-self._cellsize)
		bl.  go_to(counter.position)
		bl.end_loop()
	def op_move(self, arg, bl):
		if arg.get_type() != self.type: raise CompileError("bad type")
		self._gen_helpers()
		arg._gen_helpers()
		sdata = self._data.clone()
		odata = arg._data.clone()
		counter = self._counter.clone()
		bl.shift(self._cellsize)
		sdata.op_move(odata, bl)
		bl.go_add(counter.position, self.type.get_len() - 1)
		bl.start_loop()
		bl.  ext_moveloop(counter.position, [counter.position + self._cellsize])
		bl.  go_add(counter.position, 1)
		bl.  shift(self._cellsize)
		sdata.op_move(odata, bl)
		bl.  go_sub(counter.position, 1)
		bl.end_loop()
		bl.go_add(counter.position, 1)
		bl.start_loop()
		bl.  go_sub(counter.position, 1)
		bl.  shift(-self._cellsize)
		bl.  go_to(counter.position)
		bl.end_loop()
	def op_index(self, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		self._gen_helpers()
		holder = self._holder
		pholder = self._pholder
		data = self._data
		counter = self._counter.clone()
		counter.set_position(counter.get_position() + self._cellsize)
		counter.op_move(arg, bl)
		counter = self._counter.clone()
		bl.shift(self._cellsize)
		bl.go_to(counter.position)
		bl.start_loop()
		bl.  ext_moveloop(counter.position, [counter.position + self._cellsize])
		bl.  go_add(counter.position, 1)
		bl.  shift(self._cellsize)
		bl.  go_sub(counter.position, 1)
		bl.end_loop()
		holder.op_copy(data, bl)
		bl.go_add(counter.position, 1)
		bl.start_loop()
		bl.  go_sub(counter.position, 1)
		pholder.op_move(holder, bl)
		bl.  shift(-self._cellsize)
		bl.  go_to(counter.position)
		bl.end_loop()
		self.op_clear(bl)
		return holder
	def op_index_assign(self, sdata, arg, bl):
		if arg.get_type() != ByteType(): raise CompileError("bad type")
		if sdata.get_type() != self.type.get_vtype(): raise CompileError("bad type")
		self._gen_helpers()
		holder = self._holder
		pholder = self._pholder
		data = self._data
		counter = self._counter.clone()
		counter.set_position(counter.get_position() + self._cellsize)
		counter.op_move(arg, bl)
		counter = self._counter.clone()
		holder.op_move(sdata, bl)
		bl.shift(self._cellsize)
		bl.go_to(counter.position)
		bl.start_loop()
		holder.op_move(pholder, bl)
		bl.  ext_moveloop(counter.position, [counter.position + self._cellsize])
		bl.  go_add(counter.position, 1)
		bl.  shift(self._cellsize)
		bl.  go_sub(counter.position, 1)
		bl.end_loop()
		data.op_move(pholder, bl)
		bl.go_add(counter.position, 1)
		bl.start_loop()
		bl.  go_sub(counter.position, 1)
		bl.  shift(-self._cellsize)
		bl.  go_to(counter.position)
		bl.end_loop()

class NamingScope(object):
	def __init__(self, top, parent = None):
		self.parent = parent
		self.bottom = self.top = top
		self.vars = {}
		if parent != None: self.vars = parent.vars.copy()
		self.ownvars = []
	def get_parent(self): return self.parent
	def get_ownvars(self): return self.ownvars
	def get_bottom(self): return self.bottom
	def get_top(self): return self.top
	def __getitem__(self, var): return self.vars[var]
	def __setitem__(self, var, val): 
		if self.vars.has_key(var):
			raise CompileException("trying to mask variable")
		val.set_position(self.top)
		self.top += val.get_size()
		self.vars[var] = val
		self.ownvars.append(var)

class BlockScope(NamingScope): pass

class ArgumentScope(NamingScope):
	def __setitem__(self, var, val):
		self.bottom -= val.get_size()
		val.set_position(self.bottom)
		self.vars[var] = val

class Function(object):
	def __init__(self, fno, args, rettype):
		self.fun_no = fno
		self.ascope = ArgumentScope(-6)
		for var, val, inout in args:
			self.ascope[var] = val.create()
		self.ascope["result"] = rettype.create()
		self.scopes = [NamingScope(3, self.ascope)]
		self.temps = []
		self.temp_size = 0
		self.blocks = []
		self.args = args
		self.rettype = rettype
	def get_fun_no(self):
		return self.fun_no
	def get_arg_count(self):
		return len(self.args)
	def get_arg_type(self, n):
		return self.args[n][1]
	def is_arg_inout(self, n):
		return self.args[n][2]
	def get_ret_type(self):
		return self.rettype
	def get_top_scope(self):
		return self.scopes[-1]
	def get_stack_top(self):
		return self.get_top_scope().get_top() + self.temp_size
	def new_scope(self):
		if len(self.temps) != 0:
			raise CompileError("creating scope with temps")
		top = self.get_top_scope()
		ret = BlockScope(top.get_top(), top)
		self.scopes.append(ret)
		return ret
	def pop_scope(self):
		if len(self.temps) != 0:
			raise CompileError("popping scope with temps")
		self.scopes.pop()
	def push_temp(self, var):
		stop = self.get_stack_top()
		var.set_position(stop)
		self.temp_size += var.get_size()
		self.temps.append(var)
	def put_temp(self, var):
		stop = self.get_stack_top()
		if var.get_position() != stop:
			raise CompileError("putting temp not on top")
		self.temp_size += var.get_size()
		self.temps.append(var)
	def get_temp(self):
		if len(self.temps) == 0:
			raise CompileError("getting unexistant temp")
		return self.temps[-1]
	def pop_temp(self):
		if len(self.temps) == 0:
			raise CompileError("popping unexistant temp")
		ret = self.temps.pop()
		self.temp_size -= ret.get_size()
		return ret
	def num_temps(self):
		return len(self.temps)

class Compiler(object):
	def compile(self, parse_tree):
		self.max_fno = 1
		self.fun_map = {}
		self.functions = []
	
		if type(parse_tree) != parser.Source:
			raise CompileError("Source expected")
	
		for topst in parse_tree.children:
			if type(topst) == parser.Function:
				cargs = []
				for arg in topst.argument_list:
					cargs.append((arg.name, self.compile_type(arg.type), arg.inout))
				rettype = self.compile_type(topst.return_type)
				cfun = Function(self.max_fno, cargs, rettype)
				self.fun_map[topst.name] = cfun
				self.max_fno += 1

		if not self.fun_map.has_key("main"):
			raise CompileError("No main function")

		for topst in parse_tree.children:
			if type(topst) == parser.Function:
				self.functions.append(
					"<<<<" + self.compile_function(topst) + "<<")
			else:
				raise CompileError()

		code = ">>>+>>>" + "+" * self.fun_map["main"].get_fun_no()
		code += self.gen_selector(self.functions)

		return code

	def new_cblock(self, spos = 1):
		self.blocks.append(Block(self.max_bno, self.cfun.fun_no, spos))
		self.max_bno += 1
		return self.blocks[-1]

	def switch_cblock(self, cbl):
		self.cbl = cbl

	def compile_function(self, fun):
		self.max_bno = 1
		self.blocks = []
		self.cfun = self.fun_map[fun.name]

		self.cbl = self.new_cblock()
		self.compile_block(fun.body)
		self.cbl.finish_return()

		gen_blocks = []
		for bl in self.blocks:
			gen_blocks.append(bl.get_code())
		return self.gen_selector(gen_blocks)

	def compile_block(self, cblock):
		top_scope = self.cfun.new_scope()
		for statement in cblock.children:
			self.compile_statement(statement)
		# wyczysc zmienne
		for var in top_scope.get_ownvars():
			val = top_scope[var]
			val.op_clear(self.cbl)
		self.cfun.pop_scope()

	def compile_statement(self, statement):
		top_scope = self.cfun.get_top_scope()
		if type(statement) == parser.Assignment:
			self.compile_assignment(statement)
		elif type(statement) == parser.VariableDef:
			self.compile_variabledef(statement)
		elif type(statement) == parser.Return:
			self.compile_return(statement)
		elif type(statement) == parser.Execute:
			self.compile_expression(statement.expression)
			temp = self.cfun.pop_temp()
			temp.op_clear(self.cbl)
		elif type(statement) == parser.IfElse:
			self.compile_ifelse(statement)
		elif type(statement) == parser.While:
			self.compile_while(statement)
		elif type(statement) == parser.CodeBlock:
			self.compile_block(statement)
		else:
			raise CompileError()
		# wyczysc pozostale tempy
		for i in xrange(self.cfun.num_temps()):
			var = self.cfun.pop_temp()
			var.op_clear(self.cbl)

	def compile_return(self, ret):
		scope = self.cfun.get_top_scope()
		while type(scope) == BlockScope:
			for v in scope.get_ownvars():
				scope[v].op_clear(self.cbl)
			scope = scope.get_parent()
		self.cbl.finish_return()
		self.switch_cblock(self.new_cblock())

	def compile_variabledef(self, vardef):
		vtype = self.compile_type(vardef.type)
		top_scope = self.cfun.get_top_scope()
		top_scope[vardef.names[0]] = vtype.create()
		for n in range(1, len(vardef.names)):
			top_scope[vardef.names[n]] = vtype.create()

	def compile_ifelse(self, ifelse):
		self.compile_expression(ifelse.condition)
		cbnext = self.new_cblock()
		cbtrue = self.new_cblock()
		if ifelse.code_false: cbfalse = self.new_cblock()
		else: cbfalse = cbnext
		temp = self.cfun.pop_temp()
		if temp.get_type() != ByteType():
			raise CompileError("only bytes can be tested")
		self.cbl.go_to(0)
		self.cbl.add(cbfalse.get_block_no())
		self.cbl.go_to(temp.get_position())
		self.cbl.start_loop()
		self.cbl.start_loop()
		self.cbl.sub(1)
		self.cbl.end_loop()
		self.cbl.go_to(0)
		self.cbl.add(cbtrue.get_block_no() - cbfalse.get_block_no())
		self.cbl.go_to(temp.get_position())
		self.cbl.end_loop()
		self.cbl.finish_return()
		self.switch_cblock(cbtrue)
		self.compile_statement(ifelse.code_true)
		self.cbl.finish_gotoblock(cbnext.get_block_no())
		if ifelse.code_false:
			self.switch_cblock(cbfalse)
			self.compile_statement(ifelse.code_false)
			self.cbl.finish_gotoblock(cbnext.get_block_no())
		self.switch_cblock(cbnext)

	def compile_while(self, lwhile):
		cbnext = self.new_cblock()
		cbwhile = self.new_cblock()
		cbbody = self.new_cblock()
		self.cbl.finish_gotoblock(cbwhile.get_block_no())
		self.switch_cblock(cbwhile)
		self.compile_expression(lwhile.condition)
		temp = self.cfun.pop_temp()
		if temp.get_type() != ByteType():
			raise CompileError("only bytes can be tested")
		self.cbl.go_to(0)
		self.cbl.add(cbnext.get_block_no())
		self.cbl.go_to(temp.get_position())
		self.cbl.start_loop()
		self.cbl.start_loop()
		self.cbl.sub(1)
		self.cbl.end_loop()
		self.cbl.go_to(0)
		self.cbl.add(cbbody.get_block_no() - cbnext.get_block_no())
		self.cbl.go_to(temp.get_position())
		self.cbl.end_loop()
		self.cbl.finish_return()
		self.switch_cblock(cbbody)
		self.compile_statement(lwhile.body)
		self.cbl.finish_gotoblock(cbwhile.get_block_no())
		self.switch_cblock(cbnext)

	def compile_assignment(self, assign):
		self.compile_expression(assign.rhs)
		temp = self.cfun.pop_temp()
		self.compile_assignfrom(assign.lhs, temp)

	def compile_assignfrom(self, lhs, temp):
		if type(lhs) == parser.Variable:
			var = self.cfun.get_top_scope()[lhs.name]
			if var.get_type() != temp.get_type():
				raise CompileError("type mismatch in assignment")
			var.op_move(temp, self.cbl)
		elif type(lhs) == parser.Output:
			if temp.get_type() != ByteType():
				raise CompileError("only bytes can be outputted")
			self.cbl.go_to(temp.get_position())
			self.cbl.output()
			temp.op_clear(self.cbl)
		elif type(lhs) == parser.Index:
			if type(lhs.children[0]) != parser.Variable or len(lhs.children) != 2:
				raise CompileError("only simple array assign supported")
			self.cfun.put_temp(temp)
			self.compile_expression(lhs.children[1])
			idx = self.cfun.pop_temp()
			self.cfun.pop_temp()
			var = self.cfun.get_top_scope()[lhs.children[0].name]
			var.op_index_assign(temp, idx, self.cbl)
		else:
			raise CompileError("bad lhs in assignment")

	binary_operators = {
		 	parser.Sum: "op_add",
			parser.Difference: "op_sub",
			parser.Product: "op_mul",
			parser.Quotient: "op_div",
			parser.Modulo: "op_mod",
			parser.Equal: "op_eq",
			parser.Inequal: "op_neq",
			parser.LessThan: "op_less",
			parser.LessEqual: "op_lesseq",
			parser.MoreThan: "op_more",
			parser.MoreEqual: "op_moreeq",
			parser.BoolAnd: "op_and",
			parser.BoolOr: "op_or",
			parser.Index: "op_index" }

	def compile_expression(self, expr):
		if type(expr) == parser.Variable:
			var = self.cfun.get_top_scope()[expr.name]
			temp = var.clone()
			self.cfun.push_temp(temp)
			temp.op_copy(var, self.cbl)
		elif isinstance(expr, parser.Literal):
			self.compile_literal(expr)
		elif type(expr) == parser.FunctionCall:
			self.compile_functioncall(expr)
		elif type(expr) == parser.Input:
			temp = ByteType().create()
			self.cfun.push_temp(temp)
			self.cbl.go_to(temp.get_position())
			self.cbl.input()
		elif self.binary_operators.has_key(type(expr)):
			fname = self.binary_operators[type(expr)]
			self.compile_leftassoc_binary_operator(expr,
				lambda lhs, rhs, cbl: lhs.__getattribute__(fname)(rhs, cbl))
		else:
			raise CompileError("invalid expression")

	def compile_leftassoc_binary_operator(self, expr, op):
		self.compile_expression(expr.children[0])
		for sexpr in expr.children[1:]:
			self.compile_expression(sexpr)
			rhs = self.cfun.pop_temp()
			lhs = self.cfun.pop_temp()
			nlhs = op(lhs, rhs, self.cbl)
			self.cfun.put_temp(nlhs)

	def compile_functioncall(self, pfun):
		if not self.fun_map.has_key(pfun.name):
			raise CompileError("calling nonexistant function")
		tfun = self.fun_map[pfun.name]
		if len(pfun.arguments) != tfun.get_arg_count():
			raise CompileError("bad number of arguments")
		retval = tfun.get_ret_type().create()
		self.cfun.push_temp(retval)
		for an in range(len(pfun.arguments)-1,-1,-1):
			self.compile_expression(pfun.arguments[an])
			if self.cfun.get_temp().get_type() != tfun.get_arg_type(an):
				raise CompileError("bad argument type")
		stop = self.cfun.get_stack_top()
		cbnext = self.new_cblock(stop+1)
		self.cbl.finish_callfunc(stop, tfun.get_fun_no(), cbnext.get_block_no())
		self.switch_cblock(cbnext)
		for an in range(len(pfun.arguments)):
			arg = self.cfun.pop_temp()
			if tfun.is_arg_inout(an):
				self.compile_assignfrom(pfun.arguments[an], arg)
			else:
				arg.op_clear(self.cbl)

	def compile_literal(self, expr):
		temp = self.compile_type(expr.type).create()
		self.cfun.push_temp(temp)
		if type(expr.type) == parser.VoidType: pass
		elif type(expr.type) == parser.ByteType:
			if expr.value <= 15:
				self.cbl.go_add(temp.get_position(), expr.value)
			else:
				v1 = 3; v2 = 4
				while v1 * v2 < expr.value:
					if v1 < v2: v1 += 1
					else: v2 += 1
				self.cbl.go_add(temp.get_position() + 1, v1)
				self.cbl.start_loop()
				self.cbl.  go_add(temp.get_position(), v2)
				self.cbl.  go_sub(temp.get_position() + 1, 1)
				self.cbl.end_loop()
				self.cbl.go_sub(temp.get_position(), v1 * v2 - expr.value)
		elif type(expr.type) == parser.ArrayType and type(expr.type.type) == parser.ByteType:
			p = temp.get_position() + temp.get_type().get_vtype().get_size()
			step = temp.get_type().get_cellsize()
			for ch in expr.value:
				p += step
				self.cbl.go_add(p, ord(ch))
		else:
			raise CompileError("invalid literal")

	def compile_type(self, t):
		if type(t) == parser.VoidType:
			return VoidType()
		elif type(t) == parser.ByteType:
			return ByteType()
		elif type(t) == parser.ArrayType:
			return ArrayType(t.len, self.compile_type(t.type))
		else:
			raise CompileError()

	def gen_selector(self, blocks):
		nblocks = len(blocks)
		ret = "[>+<"
		ret += "[-" * nblocks
		ret += "[[-]>-]"
		for i in xrange(nblocks):
			ret += ">[-"
			ret += blocks[nblocks - 1 - i]
			ret += ">]<"
			ret += "]"
		ret += "<]"
		return ret
			
