# High-Level Brainfuck
#
# Copyright (c) 2004 Marek Materzok
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import parser
import string

%%

parser HLBF:
	ignore:		'\\s+'
	token NUM:	'[0-9]+'
	token ID:	'[a-zA-Z][a-zA-Z0-9_]*'
	token END:	"$"
	token TEXT:	"([a-zA-Z0-9!?., ]|\\\\([ntr]|x[0-9a-fA-F][0-9a-fA-F]))*"

	rule program:
		{{ toplevel = [] }}
		(
			function {{ toplevel.append(function) }}
		)*
		END
		{{ return parser.Source(toplevel) }}

	rule function:
		vtype {{ rtype = vtype }}
		ID {{ fname = ID }}
		"\(" {{ args = [] }}
		(
			{{ inout = False }}
			vtype ( "&" {{ inout = True }} )?
			ID {{ args.append(parser.Argument(vtype, ID, inout)) }}
			(
				{{ inout = False }}
				","
				vtype ( "&" {{ inout = True }} )?
				ID {{ args.append(parser.Argument(vtype, ID, inout)) }}
			)*
		)?
		"\)"
		codeblock {{ return parser.Function(fname, args, rtype, codeblock) }}

	rule vtype:
		"void" {{ return parser.VoidType() }}
		| "byte" {{ return parser.ByteType() }}
		| "array" "\[" NUM "\]" "of" vtype {{ return parser.ArrayType(string.atoi(NUM), vtype) }}

	rule codeblock:
		"{" {{ statements = [] }}
			(
				statement ";" {{ statements.append(statement) }}
			)*
		"}" {{ return parser.CodeBlock(statements) }}

	rule statement:
		variabledef {{ return variabledef }}
		| "do" expression {{ return parser.Execute(expression) }}
		| codeblock {{ return codeblock }}
		| ifelse {{ return ifelse }}
		| whileloop {{ return whileloop }}
		| assignment {{ return assignment }}
		| "return" {{ return parser.Return() }}
	
	rule variabledef:
		"def" vtype ID {{ ids = [ID] }}
		(
			"," ID {{ ids.append(ID) }}
		)*
		{{ return parser.VariableDef(vtype, ids) }}

	rule assignment:
		(
			variable 
			( "\[" expression "\]" {{ lhs = parser.Index([variable, expression]) }}
			| {{ lhs = variable }} )
			| "out" {{ lhs = parser.Output() }}
		)
		"<-"
		expression {{ return parser.Assignment(lhs, expression) }}

	rule ifelse:
		{{ truest = falsest = None }}
		"if" "\(" expression "\)" {{ condition = expression }}
		statement {{ truest = statement }}
		( "else" statement {{ falsest = statement }} )?
		{{ return parser.IfElse(condition, truest, falsest) }}

	rule whileloop:
		"while" "\(" expression "\)" statement
		{{ return parser.While(expression, statement) }}

	rule expression:
		or_expr {{ return or_expr }}

	rule literal:
		"null" {{ return parser.NoneLiteral() }}
		| NUM {{ return parser.ByteLiteral(string.atoi(NUM)) }}
		| "\"" TEXT "\"" {{ return parser.ByteArrayLiteral(eval("'"+TEXT+"'")) }}
	
	rule funcall:
		{{ args = [] }}
		ID "\(" {{ fname = ID }}
		(
			expression {{ args.append(expression) }}
			(
				","	expression {{ args.append(expression) }}
			)*
		)?
		"\)" {{ return parser.FunctionCall(fname, args) }}

	rule or_expr:
		and_expr {{ ret = and_expr }}
		(
			"||" and_expr {{ ret = parser.BoolOr([ret, and_expr]) }}
		)* {{ return ret }}

	rule and_expr:
		eq_expr {{ ret = eq_expr }}
		(
			"&&" eq_expr {{ ret = parser.BoolAnd([ret, eq_expr]) }}
		)* {{ return ret }}

	rule eq_expr:
		rel_expr {{ ret = rel_expr }}
		(
			"=" rel_expr {{ ret = parser.Equal([ret, rel_expr]) }}
			| "!=" rel_expr {{ ret = parser.Inequal([ret, rel_expr]) }}
		)* {{ return ret }}

	rule rel_expr:
		sum_expr {{ ret = sum_expr }}
		(
			">" sum_expr {{ ret = parser.MoreThan([ret, sum_expr]) }}
			| ">=" sum_expr {{ ret = parser.MoreEqual([ret, sum_expr]) }}
			| "<" sum_expr {{ ret = parser.LessThan([ret, sum_expr]) }}
			| "<=" sum_expr {{ ret = parser.LessEqual([ret, sum_expr]) }}
		)* {{ return ret }}

	rule sum_expr:
		factor_expr {{ ret = factor_expr }}
		(
			"\+" factor_expr {{ ret = parser.Sum([ret, factor_expr]) }}
			| "-" factor_expr {{ ret = parser.Difference([ret, factor_expr]) }}
		)* {{ return ret }}

	rule factor_expr:
		postfix_expr {{ ret = postfix_expr }}
		(
			"\*" postfix_expr {{ ret = parser.Product([ret, postfix_expr]) }}
			| "/" postfix_expr {{ ret = parser.Quotient([ret, postfix_expr]) }}
			| "%" postfix_expr {{ ret = parser.Modulo([ret, postfix_expr]) }}
		)* {{ return ret }}

	rule postfix_expr:
		term_expr {{ ret = term_expr }}
		(
			"\[" expression "\]" {{ ret = parser.Index([ret, expression]) }}
		)* {{ return ret }}

	rule term_expr:
		"in" {{ return parser.Input() }}
		| literal {{ return literal }}
		| funcall {{ return funcall }}
		| variable {{ return variable }}
		| "\(" expression "\)" {{ return expression }}

	rule variable:
		"\$" ID {{ return parser.Variable(ID) }}

